import itertools
from collections import defaultdict
from typing import List, Dict

import networkx as nx


class BlockMatcherBase:
    """
    This class takes two graphs and returns a dictionary of likley matches between the nodes in the two graphs.
    The key is nodes in g1 and the value is the corresponding node in g2. This class is expected to be used
    with two graphs you _know_ are implementing the same logic. This can happen in instances of a function being
    compiled with different optimization levels, or with different compilers.

    This implementation is based on the paper: "discovRE: Efficient Cross-Architecture Identification of Bugs
    in Binary Code" by Eschweiler et al. in NDSS 2016.
    """
    W_NO_ARITHMETIC_INS = 56.658
    W_NO_CALLS = 87.423
    W_NO_INS = 40.423
    W_NO_LOGIC_INS = 76.694
    W_NO_BRANCH_INS = 6.841
    W_STR_CONST = 11.998
    W_NUM_CONST = 15.382

    TYP_NO_ARITHMETIC_INS = "arith_ins"
    TYP_NO_CALLS = "calls"
    TYP_NO_INS = "ins"
    TYP_NO_LOGIC_INS = "logic_ins"
    TYP_NO_BRANCH_INS = "branch_ins"
    TYP_STR_CONST = "str_consts"
    TYP_NUM_CONST = "num_consts"

    def __init__(self, g1: nx.DiGraph, g2: nx.DiGraph):
        self._g1 = g1
        self._g2 = g2
        self._g1_nodes = list(g1.nodes)
        self._g2_nodes = list(g2.nodes)

        self.mapping = {}

    def analyze(self):
        # TODO: add a way to first match nodes with exact call matches
        # TODO: add a way to match nodes with exact long string matches
        sim_map = self.generate_similarities()
        for b1, b2_scores in sim_map.items():
            # TODO: what happens if b1->b2 score is lower than another node could match to in b1?
            best_match = min(b2_scores, key=b2_scores.get)
            self.mapping[b1] = best_match

    def generate_similarities(self) -> Dict[object, Dict[object, float]]:
        scores = defaultdict(dict)
        for b1, b2 in itertools.product(self._g1_nodes, self._g2_nodes):
            scores[b1][b2] = self.block_similarity(b1, b2, self._g1, self._g2)

        return scores

    def block_similarity(self, b1, b2, g1, g2) -> float:
        feat_func_map = {
            self.get_number_of_arithmetic_ins: self.W_NO_ARITHMETIC_INS,
            self.get_number_of_calls: self.W_NO_CALLS,
            self.get_number_of_ins: self.W_NO_INS,
            self.get_number_of_logic_ins: self.W_NO_LOGIC_INS,
            self.get_number_of_branch_ins: self.W_NO_BRANCH_INS,
        }
        listed_feat_func_map = {
            self.get_str_consts: self.W_STR_CONST,
            self.get_num_consts: self.W_NUM_CONST
        }

        numerator = 0
        denominator = 0
        for feat_func, weight in feat_func_map.items():
            f1, f2 = feat_func(b1, g1), feat_func(b2, g2)
            numerator += weight * abs(f1 - f2)
            denominator += weight * max(f1, f2)

        # special case for lists of features
        for feat_func, weight in listed_feat_func_map.items():
            f1, f2 = feat_func(b1, g1), feat_func(b2, g2)
            # jaccard similarity
            numerator += (len(set.union(set(f1), set(f2))) - len(set.intersection(set(f1), set(f2)))) * weight
            denominator += len(set.union(set(f1), set(f2))) * weight

        # denominator == 0 when two empty blocks compared, they are equal
        return (numerator / denominator) if denominator else 0

    #
    # Must be implemented by the user
    #

    def get_number_of_arithmetic_ins(self, block, graph) -> int:
        raise NotImplementedError()

    def get_number_of_calls(self, block, graph) -> int:
        raise NotImplementedError()

    def get_number_of_ins(self, block, graph) -> int:
        raise NotImplementedError()

    def get_number_of_logic_ins(self, block, graph) -> int:
        raise NotImplementedError()

    def get_number_of_branch_ins(self, block, graph) -> int:
        raise NotImplementedError()

    def get_str_consts(self, block, graph) -> List[str]:
        raise NotImplementedError()

    def get_num_consts(self, block, graph) -> List[int]:
        raise NotImplementedError()



from collections import defaultdict
from typing import List

import networkx as nx
from ailment import Block
from ailment.statement import Call

from cfgutils.sorting import quasi_topological_sort_nodes
from cfgutils.similarity.block_matcher_base import BlockMatcherBase
from .feat_extractor import AILBlockFeatureExtractor


class AILBlockMatcher(BlockMatcherBase):
    def __init__(self, g1: nx.DiGraph, g2: nx.DiGraph, proj1=None, proj2=None):
        super().__init__(g1, g2)
        self._proj1 = proj1
        self._proj2 = proj2
        self._g1_cache = defaultdict(dict)
        self._g2_cache = defaultdict(dict)
        self.generate_feat_cache()
        self.analyze()

    def generate_feat_cache(self):
        for g, proj, features in [(self._g1, self._proj1, self._g1_cache), (self._g2, self._proj2, self._g2_cache)]:
            proj_cfg = proj.kb.cfgs.get_most_accurate() if proj is not None else None
            for node in g.nodes:
                feat_extractor = AILBlockFeatureExtractor(proj, proj_cfg)
                feat_extractor.walk(node)
                features[node][self.TYP_NO_ARITHMETIC_INS] = len(feat_extractor.arith_ins)
                features[node][self.TYP_NO_CALLS] = len(feat_extractor.calls)
                features[node][self.TYP_NO_INS] = len(feat_extractor.ins)
                features[node][self.TYP_NO_LOGIC_INS] = len(feat_extractor.logic_ins)
                features[node][self.TYP_NO_BRANCH_INS] = len(feat_extractor.branch_ins)
                features[node][self.TYP_STR_CONST] = feat_extractor.str_consts
                features[node][self.TYP_NUM_CONST] = feat_extractor.num_consts

    def _get_correct_cache(self, graph):
        if graph == self._g1:
            return self._g1_cache
        elif graph == self._g2:
            return self._g2_cache
        else:
            raise ValueError("Graph not found")

    def get_number_of_arithmetic_ins(self, block, graph) -> int:
        cache = self._get_correct_cache(graph)
        return cache[block][self.TYP_NO_ARITHMETIC_INS]

    def get_number_of_calls(self, block, graph) -> int:
        cache = self._get_correct_cache(graph)
        return cache[block][self.TYP_NO_CALLS]

    def get_number_of_ins(self, block, graph) -> int:
        cache = self._get_correct_cache(graph)
        return cache[block][self.TYP_NO_INS]

    def get_number_of_logic_ins(self, block, graph) -> int:
        cache = self._get_correct_cache(graph)
        return cache[block][self.TYP_NO_LOGIC_INS]

    def get_number_of_branch_ins(self, block, graph) -> int:
        cache = self._get_correct_cache(graph)
        return cache[block][self.TYP_NO_BRANCH_INS]

    def get_str_consts(self, block, graph) -> List[str]:
        cache = self._get_correct_cache(graph)
        return cache[block][self.TYP_STR_CONST]

    def get_num_consts(self, block, graph) -> List[int]:
        cache = self._get_correct_cache(graph)
        return cache[block][self.TYP_NUM_CONST]



from typing import Optional, Any

from ailment import BinaryOp, UnaryOp
from ailment.block_walker import AILBlockWalkerBase
from ailment.block import Block
from ailment.statement import Call, ConditionalJump, Statement
from ailment.expression import Const

from .prettyify_ail import string_at_addr


class AILBlockFeatureExtractor(AILBlockWalkerBase):
    """
    Extracts features from AIL blocks based on discoveRE features
    """
    LOGIC_INS_TYPS = {
        "LogicalAnd", "LogicalOr", "CmpF", "CmpEQ", "CmpNE", "CmpLT", "CmpLE", "CmpGT", "CmpGE", "CmpLTs",
        "CmpLEs", "CmpGTs", "CmpGEs"
    }
    ARITH_INS_TYPS = {
        "Add", "AddF", "Sub", "SubF", "Mul", "MulF", "Div", "DivF", "DivMo", "Mod", "Xor", "And",
        "Or", "Shl", "Shr", "Sar", "Ror", "Rol", "Not"
    }

    def __init__(self, project=None, project_cfg=None):
        self._project = project
        self._project_cfg = project_cfg

        # features
        self.arith_ins = []
        self.calls = []
        self.ins = []
        self.logic_ins = []
        self.branch_ins = []
        self.str_consts = []
        self.num_consts = []

        super().__init__()

    def _handle_stmt(self, stmt_idx: int, stmt: Statement, block: Optional[Block]) -> Any:
        self.ins.append(stmt)
        return super()._handle_stmt(stmt_idx, stmt, block)

    def _handle_BinaryOp(self, expr_idx: int, expr: BinaryOp, stmt_idx: int, stmt: Statement, block: Optional[Block]):
        if expr.op in self.ARITH_INS_TYPS:
            self.arith_ins.append(expr)
        elif expr.op in self.LOGIC_INS_TYPS:
            self.logic_ins.append(expr)
        return super()._handle_BinaryOp(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_UnaryOp(self, expr_idx: int, expr: UnaryOp, stmt_idx: int, stmt: Statement, block: Optional[Block]):
        if expr.op in self.ARITH_INS_TYPS:
            self.logic_ins.append(expr)
        return super()._handle_UnaryOp(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_ConditionalJump(self, stmt_idx: int, stmt: ConditionalJump, block: Optional[Block]):
        self.branch_ins.append(stmt)
        return super()._handle_ConditionalJump(stmt_idx, stmt, block)

    def _handle_CallExpr(self, expr_idx: int, expr: "Call", stmt_idx: int, stmt, block: Optional["Block"]):
        self.calls.append(expr)
        super()._handle_CallExpr(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_Call(self, stmt_idx: int, stmt: "Call", block: Optional["Block"]):
        self.calls.append(stmt)
        super()._handle_Call(stmt_idx, stmt, block)

    def _handle_Const(self, expr_idx: int, expr: "Const", stmt_idx: int, stmt: Statement, block: Optional[Block]):
        if self._project is not None and self._project_cfg is not None:
            str_val = string_at_addr(self._project_cfg, expr.value, self._project, max_size=200)
            if str_val is not None:
                self.str_consts.append(str_val)
            else:
                self.num_consts.append(expr.value)
        else:
            self.num_consts.append(expr.value)

        return super()._handle_Const(expr_idx, expr, stmt_idx, stmt, block)




import pickle
from pathlib import Path
import logging
from typing import Union, Dict, Tuple

import ailment
import networkx as nx
import angr
from angr.analyses.decompiler.optimization_passes import DUPLICATING_OPTS, CONDENSING_OPTS
from angr.analyses.decompiler.utils import to_ail_supergraph

from .prettyify_ail import stmt_to_pretty_text
from cfgutils.data.generic_block import GenericBlock

_l = logging.getLogger(__name__)


def binary_to_ail_cfgs(
    binary_path: Path, functions=None, make_generic=False, structuring_opts=True, supergraph=True,
    return_project=False,
) -> Union[Dict[str, nx.DiGraph], Tuple[Dict[str, nx.DiGraph], angr.Project]]:
    """
    A simple wrapper around the angr decompiler to simply use the defaults and return the AIL CFGs which
    are present at the end of the decompilation process. Using make_generic, you can convert the AIL CFGs
    into GenericBlock CFGs for use with CFGUtils functions. You can also disable some of the structuring
    optimizations if you want to see the raw CFGs.

    :param binary_path: Path to the binary to decompile
    :param functions: List of function addresses to decompile. If None, all functions will be decompiled
    :param make_generic: Convert the AIL CFGs to GenericBlock CFGs
    :param structuring_opts: Enable or disable structuring optimizations
    :param supergraph: Convert the AIL CFGs to supergraphs
    :param return_project: Return the angr Project object as well
    """
    binary_path = Path(binary_path).absolute()
    if not binary_path.exists():
        raise FileNotFoundError(f"{binary_path} does not exist")

    proj = angr.Project(binary_path, auto_load_libs=False)
    cfg = proj.analyses.CFG(show_progressbar=False, normalize=True, data_references=True)
    try:
        proj.analyses.CompleteCallingConventions(cfg=cfg, recover_variables=True, analyze_callsites=True)
        cc_failed = False
    except Exception:
        cc_failed = True

    if cc_failed:
        _l.warning(f"CallingConvention Analysis failed on {binary_path}. Trying again without variable recovery...")
        try:
            # try it again without variable recovery
            proj.analyses.CompleteCallingConventions(cfg=cfg, recover_variables=False)
            cc_failed = False
        except Exception:
            pass

    if cc_failed:
        _l.critical(f"All attempts to run CallingConvention Analysis failed on {binary_path}.")

    # clean up function names
    functions = functions or cfg.functions
    for func in list(cfg.functions.values()):
        if "." in func.name:
            func.name = func.name[:func.name.index(".")]

    # some optimizations can drastically change the structure of the CFG, so we should disable some if wanted
    all_optimizations = angr.analyses.decompiler.optimization_passes.get_default_optimization_passes(
        proj.arch, "linux"
    )
    if not structuring_opts:
        all_optimizations = [opt for opt in all_optimizations if opt not in DUPLICATING_OPTS + CONDENSING_OPTS]

    # generate a cfg for each function
    ail_cfgs = {}
    for func_idx, func_addr in enumerate(functions):
        try:
            f = cfg.functions[func_addr]
        except Exception:
            _l.critical(f"Function at {func_addr} not found in CFG")
            continue

        if f is None or f.is_plt:
            continue

        # for this function you don't actually need the linear decompilation, but we run through the entire
        # decompilation process to assure every optimization is run that would be done on a normal Clinic graph
        dec = proj.analyses.Decompiler(f, cfg=cfg, kb=cfg.kb, optimization_passes=all_optimizations)
        dec.clinic.cc_graph.name = str(f.name)
        ail_cfgs[str(f.name)] = dec.clinic.cc_graph if not supergraph else to_ail_supergraph(dec.clinic.cc_graph)

    if make_generic:
        cfgs = [ail_cfg_to_generic(cfg, proj) for name, cfg in ail_cfgs.items()]
    else:
        cfgs = list(ail_cfgs.values())

    named_cfgs = {cfg.name or str(i): cfg for i, cfg in enumerate(cfgs)}
    if return_project:
        return named_cfgs, proj
    else:
        return named_cfgs


def ail_cfg_to_generic(cfg: nx.DiGraph, project=None):
    new_cfg = nx.DiGraph()
    new_cfg.name = cfg.name
    proj_cfg = project.kb.cfgs.get_most_accurate() if project is not None else None
    node_map = {}
    for node in cfg.nodes:
        new_node = GenericBlock(node.addr, idx=node.idx if hasattr(node, "idx") else None)
        for stmt in node.statements:
            str_stmt = stmt_to_pretty_text(stmt, project, proj_cfg) if project is not None else str(stmt)
            new_node.statements.append(str_stmt)

        new_cfg.add_node(new_node, node=new_node)
        node_map[node] = new_node

    for src, dst in cfg.edges:
        new_src = node_map[src]
        new_dst = node_map[dst]
        new_cfg.add_edge(new_src, new_dst, src=new_src, dst=new_dst)

    return new_cfg


def binary_to_generic_cfgs(binary_path: Path, functions=None):
    return binary_to_ail_cfgs(binary_path, functions, make_generic=True)


def ail_pickle_to_cfg(pickle_path: Path):
    """
    This function converts a pickle files, containing an AIL CFG, into a GenericBlock CFG for use with
    CFGUtils functions. Note: this pickled AIL CFG is a little special. All the nodes in the CFG are just
    the addresses of the blocks (not real AIL nodes). The 'label' attribute of each node is a list of strings,
    which is simply each AIL stmt stringed.
    """

    with open(pickle_path, "rb") as fp:
        ail_str_cfg: nx.DiGraph = pickle.load(fp)

    new_cfg = nx.DiGraph()
    node_map = {}
    for node in ail_str_cfg.nodes:
        new_node = GenericBlock(int(node), statements=ail_str_cfg.nodes[node]["label"])
        new_cfg.add_node(new_node, node=new_node)
        node_map[node] = new_node

    for src, dst in ail_str_cfg.edges:
        new_src = node_map[src]
        new_dst = node_map[dst]
        new_cfg.add_edge(new_src, new_dst, src=new_src, dst=new_dst)

    return new_cfg




from cfgutils.similarity.ged.abu_aisheh_ged import (
    ged_exact, ged_max, ged_upperbound, ged_explained, graph_edit_distance_core_analysis
)
from cfgutils.similarity.ged.basque_cfged import cfg_edit_distance
from cfgutils.similarity.ged.hu_cfged import hu_cfged
from cfgutils.similarity.block_matcher_base import BlockMatcherBase
